@@ -85,9 +85,14 @@ Pragma to the header of your file.
 -}
 
 {-# LANGUAGE CPP           #-}
+
+{-# LANGUAGE BangPatterns  #-}
+{-# LANGUAGE DataKinds     #-}
 {-# LANGUAGE LambdaCase    #-}
+{-# LANGUAGE MultiWayIf    #-}
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE ViewPatterns  #-}
+{-# LANGUAGE TemplateHaskellQuotes #-}
 
 {-# LANGUAGE Trustworthy   #-}
 
@@ -97,93 +102,87 @@ module GHC.TypeLits.KnownNat.Solver
   ( plugin )
 where
 
--- external
-import Control.Arrow                ((&&&), first)
-import Control.Monad.Trans.Maybe    (MaybeT (..))
+-- base
+import Control.Arrow
+  ( (&&&), first )
+import Data.Foldable
+  ( asum )
+import Data.List.NonEmpty as NE
+  ( filter )
+import Data.Maybe
+  ( catMaybes, fromMaybe, mapMaybe )
+
+-- transformers
+import Control.Monad.Trans.Maybe
+  ( MaybeT (..) )
 import Control.Monad.Trans.Writer.Strict
-import Data.Maybe                   (catMaybes,mapMaybe)
-import GHC.TcPluginM.Extra          (lookupModule, lookupName, newWanted,
-                                     tracePlugin)
-#if MIN_VERSION_ghc(8,4,0)
-import GHC.TcPluginM.Extra          (flattenGivens, mkSubst', substType)
-#endif
-import GHC.TypeLits.Normalise.SOP   (SOP (..), Product (..), Symbol (..))
-import GHC.TypeLits.Normalise.Unify (CType (..),normaliseNat,reifySOP)
 
--- GHC API
-import Class      (Class, classMethods, className, classTyCon)
-#if MIN_VERSION_ghc(8,6,0)
-import Coercion   (Role (Representational), mkUnivCo)
-#endif
-import FamInst    (tcInstNewTyCon_maybe)
-import FastString (fsLit)
-import Id         (idType)
-import InstEnv    (instanceDFunId,lookupUniqueInstEnv)
-#if MIN_VERSION_ghc(8,5,0)
-import MkCore     (mkNaturalExpr)
-#endif
-import Module     (mkModuleName, moduleName, moduleNameString)
-import Name       (nameModule_maybe, nameOccName)
-import OccName    (mkTcOcc, occNameString)
-import Plugins    (Plugin (..), defaultPlugin)
-#if MIN_VERSION_ghc(8,6,0)
-import Plugins    (purePlugin)
-#endif
-import PrelNames  (knownNatClassName)
-#if MIN_VERSION_ghc(8,5,0)
-import TcEvidence (EvTerm (..), EvExpr, evDFunApp, mkEvCast, mkTcSymCo, mkTcTransCo)
-#else
-import TcEvidence (EvTerm (..), EvLit (EvNum), mkEvCast, mkTcSymCo, mkTcTransCo)
-#endif
-#if MIN_VERSION_ghc(8,5,0)
-import TcPluginM  (unsafeTcPluginTcM)
-#endif
-#if !MIN_VERSION_ghc(8,4,0)
-import TcPluginM  (zonkCt)
-#endif
-import TcPluginM  (TcPluginM, tcLookupClass, getInstEnvs)
-import TcRnTypes  (TcPlugin(..), TcPluginResult (..))
-import TcTypeNats (typeNatAddTyCon, typeNatSubTyCon)
-#if MIN_VERSION_ghc(8,4,0)
-import TcTypeNats (typeNatDivTyCon)
-#endif
-import Type
-  (PredType,
-   dropForAlls, eqType, funResultTy, mkNumLitTy, mkStrLitTy, mkTyConApp,
-   piResultTys, splitFunTys, splitTyConApp_maybe, tyConAppTyCon_maybe, typeKind)
-import TyCon      (tyConName)
-import TyCoRep    (Type (..), TyLit (..))
-#if MIN_VERSION_ghc(8,6,0)
-import TyCoRep    (UnivCoProvenance (PluginProv))
-import TysWiredIn (boolTy)
-#endif
-import Var        (DFunId)
+-- ghc-typelits-natnormalise
+import GHC.TypeLits.Normalise.SOP
+  ( SOP (..), Product (..), Symbol (..) )
+import GHC.TypeLits.Normalise.Unify
+  ( CType (..),normaliseNat, reifySOP )
 
-#if MIN_VERSION_ghc(8,10,0)
-import Constraint
-  (Ct, ctEvExpr, ctEvidence, ctEvLoc, ctEvPred, ctLoc, ctLocSpan, isWanted,
-   mkNonCanonical, setCtLoc, setCtLocSpan)
-import Predicate (EqRel (NomEq), Pred (ClassPred,EqPred), classifyPredType)
-#else
-import TcRnTypes
-  (Ct, ctEvidence, ctEvLoc, ctEvPred, ctLoc, ctLocSpan, isWanted, mkNonCanonical,
-   setCtLoc, setCtLocSpan)
-import Type      (EqRel (NomEq), PredTree (ClassPred,EqPred), classifyPredType)
-#if MIN_VERSION_ghc(8,5,0)
-import TcRnTypes (ctEvExpr)
-#else
-import TcRnTypes (ctEvTerm)
-#endif
-#endif
+-- ghc-tcplugin-api
+import GHC.TcPlugin.API
+import GHC.TcPlugin.API.TyConSubst
 
--- | Classes and instances from "GHC.TypeLits.KnownNat"
-data KnownNatDefs
-  = KnownNatDefs
-  { knownBool     :: Class
-  , knownBoolNat2 :: Class
-  , knownNat2Bool :: Class
-  , knownNatN     :: Int -> Maybe Class -- ^ KnownNat{N}
-  }
+-- ghc-typelits-knownnat
+import GHC.TypeLits.KnownNat.Compat
+  ( KnownNatDefs(..), lookupKnownNatDefs, mkNaturalExpr
+  , coercionRKind, classMethodTy
+  , irrelevantMult
+  )
+
+-- ghc
+import GHC.Builtin.Names
+  ( knownNatClassName )
+#if MIN_VERSION_ghc(9,1,0)
+import GHC.Builtin.Types
+  ( promotedFalseDataCon, promotedTrueDataCon )
+import GHC.Builtin.Types.Literals
+  ( typeNatCmpTyCon )
+#endif
+import GHC.Builtin.Types.Literals
+  ( typeNatAddTyCon, typeNatDivTyCon, typeNatSubTyCon )
+import GHC.Core
+  ( mkApps, mkTyApps )
+import GHC.Core.Class
+  ( classMethods, classTyVars )
+import GHC.Core.Coercion
+  ( instNewTyCon_maybe, mkNomReflCo, mkTyConAppCo )
+import GHC.Core.DataCon
+  ( dataConWrapId )
+import GHC.Core.InstEnv
+  ( instanceDFunId, lookupUniqueInstEnv )
+import GHC.Core.TyCo.Rep
+  ( Type(..), TyLit(..) )
+import GHC.Core.TyCo.Subst
+  ( substTyWithUnchecked )
+import GHC.Core.Type
+  ( coreView, piResultTys, splitFunTys )
+import GHC.Core.Utils
+  ( exprType, mkCast )
+import GHC.Driver.Plugins
+  ( Plugin (..), defaultPlugin, purePlugin )
+import GHC.Plugins
+  ( HasDebugCallStack )
+import GHC.Tc.Types.Evidence
+  ( evTermCoercion_maybe, evSelector )
+import GHC.Types.Id
+  ( idType )
+import GHC.Types.Name
+  ( nameModule_maybe, nameOccName )
+import GHC.Types.Name.Occurrence
+  ( occNameString )
+import GHC.Types.Var
+  ( DFunId )
+import GHC.Unit.Module
+  ( moduleName, moduleNameString )
+import GHC.Utils.Outputable
+  ( (<+>), vcat, text )
+
+--------------------------------------------------------------------------------
 
 -- | Simple newtype wrapper to distinguish the original (flattened) argument of
 -- knownnat from the un-flattened version that we work with internally.
@@ -281,45 +280,34 @@ Pragma to the header of your file.
 plugin :: Plugin
 plugin
   = defaultPlugin
-  { tcPlugin = const $ Just normalisePlugin
-#if MIN_VERSION_ghc(8,6,0)
+  { tcPlugin = \ _ -> Just $ mkTcPlugin normalisePlugin
   , pluginRecompile = purePlugin
-#endif
   }
 
 normalisePlugin :: TcPlugin
-normalisePlugin = tracePlugin "ghc-typelits-knownnat"
+normalisePlugin =
   TcPlugin { tcPluginInit  = lookupKnownNatDefs
            , tcPluginSolve = solveKnownNat
+           , tcPluginRewrite = const emptyUFM
            , tcPluginStop  = const (return ())
            }
 
-solveKnownNat :: KnownNatDefs -> [Ct] -> [Ct] -> [Ct]
-              -> TcPluginM TcPluginResult
-solveKnownNat _defs _givens _deriveds []      = return (TcPluginOk [] [])
-solveKnownNat defs  givens  _deriveds wanteds = do
-  -- GHC 7.10 puts deriveds with the wanteds, so filter them out
-  let wanteds'   = filter (isWanted . ctEvidence) wanteds
-#if MIN_VERSION_ghc(8,4,0)
-      subst      = map fst
-                 $ mkSubst' givens
-      kn_wanteds = map (\(x,y,z,orig) -> (x,y,substType subst z,orig))
-                 $ mapMaybe (toKnConstraint defs) wanteds'
-#else
-      kn_wanteds = mapMaybe (toKnConstraint defs) wanteds'
-#endif
+solveKnownNat :: KnownNatDefs -> [Ct] -> [Ct]
+              -> TcPluginM Solve TcPluginSolveResult
+solveKnownNat _defs _givens []      = return (TcPluginOk [] [])
+solveKnownNat defs  givens  wanteds = do
+  let givensTyConSubst = mkTyConSubst givens
+      kn_wanteds = map (\(x,y,z,orig) -> (x,y,z,orig))
+                 $ mapMaybe (toKnConstraint defs) wanteds
   case kn_wanteds of
     [] -> return (TcPluginOk [] [])
     _  -> do
       -- Make a lookup table for all the [G]iven constraints
-#if MIN_VERSION_ghc(8,4,0)
-      let given_map = map toGivenEntry (flattenGivens givens)
-#else
-      given_map <- mapM (fmap toGivenEntry . zonkCt) givens
-#endif
+      let given_map = map toGivenEntry givens
+
       -- Try to solve the wanted KnownNat constraints given the [G]iven
       -- KnownNat constraints
-      (solved,new) <- (unzip . catMaybes) <$> (mapM (constraintToEvTerm defs given_map) kn_wanteds)
+      (solved,new) <- (unzip . catMaybes) <$> (mapM (constraintToEvTerm defs givensTyConSubst given_map) kn_wanteds)
       return (TcPluginOk solved (concat new))
 
 -- | Get the KnownNat constraints
@@ -332,61 +320,22 @@ toKnConstraint defs ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
   _ -> Nothing
 
 -- | Create a look-up entry for a [G]iven constraint.
-#if MIN_VERSION_ghc(8,5,0)
 toGivenEntry :: Ct -> (CType,EvExpr)
-#else
-toGivenEntry :: Ct -> (CType,EvTerm)
-#endif
 toGivenEntry ct = let ct_ev = ctEvidence ct
                       c_ty  = ctEvPred   ct_ev
-#if MIN_VERSION_ghc(8,5,0)
                       ev    = ctEvExpr   ct_ev
-#else
-                      ev    = ctEvTerm   ct_ev
-#endif
                   in  (CType c_ty,ev)
 
--- | Find the \"magic\" classes and instances in "GHC.TypeLits.KnownNat"
-lookupKnownNatDefs :: TcPluginM KnownNatDefs
-lookupKnownNatDefs = do
-    md     <- lookupModule myModule myPackage
-    kbC    <- look md "KnownBool"
-    kbn2C  <- look md "KnownBoolNat2"
-    kn2bC  <- look md "KnownNat2Bool"
-    kn1C   <- look md "KnownNat1"
-    kn2C   <- look md "KnownNat2"
-    kn3C   <- look md "KnownNat3"
-    return KnownNatDefs
-           { knownBool     = kbC
-           , knownBoolNat2 = kbn2C
-           , knownNat2Bool = kn2bC
-           , knownNatN     = \case { 1 -> Just kn1C
-                                   ; 2 -> Just kn2C
-                                   ; 3 -> Just kn3C
-                                   ; _ -> Nothing
-                                   }
-           }
-  where
-    look md s = do
-      nm   <- lookupName md (mkTcOcc s)
-      tcLookupClass nm
-
-    myModule  = mkModuleName "GHC.TypeLits.KnownNat"
-    myPackage = fsLit "ghc-typelits-knownnat"
-
 -- | Try to create evidence for a wanted constraint
 constraintToEvTerm
-  :: KnownNatDefs     -- ^ The "magic" KnownNatN classes
-#if MIN_VERSION_ghc(8,5,0)
+  :: KnownNatDefs
+  -- ^ The "magic" KnownNatN classes
+  -> TyConSubst
   -> [(CType,EvExpr)]
-#else
-  -> [(CType,EvTerm)]
-#endif
-  -- All the [G]iven constraints
-
+  -- ^ All the [G]iven constraints
   -> KnConstraint
-  -> TcPluginM (Maybe ((EvTerm,Ct),[Ct]))
-constraintToEvTerm defs givens (ct,cls,op,orig) = do
+  -> TcPluginM Solve (Maybe ((EvTerm,Ct),[Ct]))
+constraintToEvTerm defs givensTyConSubst givens (ct,cls,op,orig) = do
     -- 1. Determine if we are an offset apart from a [G]iven constraint
     offsetM <- offset op
     evM     <- case offsetM of
@@ -394,15 +343,31 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
                  found@Just {} -> return found
                  -- 3.b If not, we check if the outer type-level operation
                  -- has a corresponding KnownNat<N> instance.
-                 _ -> go op
+                 _ -> go [] (op,Nothing)
     return ((first (,ct)) <$> evM)
   where
     -- Determine whether the outer type-level operation has a corresponding
     -- KnownNat<N> instance, where /N/ corresponds to the arity of the
     -- type-level operation
-    go :: Type -> TcPluginM (Maybe (EvTerm,[Ct]))
-    go (go_other -> Just ev) = return (Just (ev,[]))
-    go ty@(TyConApp tc args0)
+    go :: [Coercion] -> (Type, Maybe Coercion) -> TcPluginM Solve (Maybe (EvTerm,[Ct]))
+    -- Look through type aliases
+    go deps (coreView -> Just tyN, coM) = go deps (tyN, coM)
+    -- Look through rewrites
+    go deps0 (ty, coM)
+      | Just tcapps <- splitTyConApp_upTo givensTyConSubst ty
+      -- We are only interested in the splitTyConApp_upTo result that used a
+      -- rewrite
+      , withDeps@(_:_) <- NE.filter (\(_,_,deps) -> not (null deps)) tcapps
+      = do results <- traverse (\(tc, args, deps1) -> go (deps0 <> deps1)
+                                                         (TyConApp tc args, coM))
+                               withDeps
+           return (asum results)
+    -- See whether there is a given that matches it (after having looked through
+    -- type aliases and rewrites)
+    go deps (go_other deps -> Just ev, _) = return (Just (ev,[]))
+    -- And if there isn't, see whether we can construct it using a KnownNat<N>
+    -- instance
+    go deps (ty@(TyConApp tc args0), sM)
       | let tcNm = tyConName tc
       , Just m <- nameModule_maybe tcNm
       = do
@@ -412,23 +377,38 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
             fn0 = mS ++ "." ++ tcS
             fn1 = mkStrLitTy (fsLit fn0)
             args1 = fn1:args0
-            instM = case () of
-              () | Just knN_cls    <- knownNatN defs (length args0)
+            instM =
+              if | Just knN_cls    <- knownNatN defs (length args0)
                  , Right (inst, _) <- lookupUniqueInstEnv ienv knN_cls args1
                  -> Just (inst,knN_cls,args0,args1)
-                 | length args0 == 2
+  -- TODO: we should re-use the parsing functionality
+  -- that is in GHC.TypeLits.NatNormalise.Compat.
+#if MIN_VERSION_ghc(9,1,0)
+                 | tc == ordCondTyCon defs
+                 , [_,cmpNat,TyConApp t1 [],TyConApp t2 [],TyConApp f1 []] <- args0
+                 , TyConApp cmpNatTc args2@(arg2:_) <- cmpNat
+                 , cmpNatTc == typeNatCmpTyCon
+                 , t1 == promotedTrueDataCon
+                 , t2 == promotedTrueDataCon
+                 , f1 == promotedFalseDataCon
                  , let knN_cls = knownBoolNat2 defs
-                       ki      = typeKind (head args0)
+                       ki      = typeKind arg2
+                       args1N  = ki:fn1:args2
+                 , Right (inst,_) <- lookupUniqueInstEnv ienv knN_cls args1N
+                 -> Just (inst,knN_cls,args2,args1N)
+#endif
+                 | [arg0,_] <- args0
+                 , let knN_cls = knownBoolNat2 defs
+                       ki      = typeKind arg0
                        args1N  = ki:args1
                  , Right (inst, _) <- lookupUniqueInstEnv ienv knN_cls args1N
                  -> Just (inst,knN_cls,args0,args1N)
-                 | length args0 == 4
-                 , fn0 == "Data.Type.Bool.If"
-                 , let args0N = tail args0
-                       args1N = head args0:fn1:tail args0
+                 | (arg0:args0Rest@[_,_,_]) <- args0
+                 , tc == ifTyCon defs
+                 , let args1N = arg0:fn1:args0Rest
                        knN_cls = knownNat2Bool defs
                  , Right (inst, _) <- lookupUniqueInstEnv ienv knN_cls args1N
-                 -> Just (inst,knN_cls,args0N,args1N)
+                 -> Just (inst,knN_cls,args0Rest,args1N)
                  | otherwise
                  -> Nothing
         case instM of
@@ -439,7 +419,7 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
                         . splitFunTys          -- ([KnownNat x, KnowNat y], DKnownNat2 "+" x y)
                         . (`piResultTys` args0N) -- (KnowNat x, KnownNat y) => DKnownNat2 "+" x y
                         $ idType df_id         -- forall a b . (KnownNat a, KnownNat b) => DKnownNat2 "+" a b
-            (evs,new) <- unzip <$> mapM go_arg df_args
+            (evs,new) <- unzip <$> mapM (go_arg . irrelevantMult) df_args
             if className cls == className (knownBool defs)
                -- Create evidence using the original, flattened, argument of
                -- the KnownNat we're trying to solve. Not doing this results in
@@ -475,31 +455,23 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
                -- So while the "shape" of the KnownNat evidence that we return
                -- follows 'CLog 2 n + 1', the type of the evidence will be
                -- 'KnownNat fsk'; the one GHC originally asked us to solve.
-               then return ((,concat new) <$> makeOpDictByFiat df cls args1N args0N (unOrig orig) evs)
-               else return ((,concat new) <$> makeOpDict df cls args1N args0N (unOrig orig) evs)
-          _ -> return ((,[]) <$> go_other ty)
+               then return ((,concat new) <$> makeOpDictByFiat df cls args1N args0N (unOrig orig) deps evs)
+               else return ((,concat new) <$> makeOpDict df cls args1N args0N (unOrig orig) deps evs (fmap (ty,) sM))
+          _ -> return ((,[]) <$> go_other deps ty)
 
-    go (LitTy (NumTyLit i))
+    go deps ((LitTy (NumTyLit i)), _)
       -- Let GHC solve simple Literal constraints
       | LitTy _ <- op
       = return Nothing
       -- This plugin only solves Literal KnownNat's that needed to be normalised
       -- first
       | otherwise
-#if MIN_VERSION_ghc(8,5,0)
-      = (fmap (,[])) <$> makeLitDict cls op i
-#else
-      = return ((,[]) <$> makeLitDict cls op i)
-#endif
-    go _ = return Nothing
+      = (fmap (,[])) <$> makeLitDict cls op deps i
+    go _ _ = return Nothing
 
     -- Get EvTerm arguments for type-level operations. If they do not exist
     -- as [G]iven constraints, then generate new [W]anted constraints
-#if MIN_VERSION_ghc(8,5,0)
-    go_arg :: PredType -> TcPluginM (EvExpr,[Ct])
-#else
-    go_arg :: PredType -> TcPluginM (EvTerm,[Ct])
-#endif
+    go_arg :: PredType -> TcPluginM Solve (EvExpr,[Ct])
     go_arg ty = case lookup (CType ty) givens of
       Just ev -> return (ev,[])
       _ -> do
@@ -508,22 +480,19 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
 
     -- Fall through case: look up the normalised [W]anted constraint in the list
     -- of [G]iven constraints.
-    go_other :: Type -> Maybe EvTerm
-    go_other ty =
+    go_other :: [Coercion] -> Type -> Maybe EvTerm
+    go_other deps ty =
       let knClsTc = classTyCon cls
           kn      = mkTyConApp knClsTc [ty]
           cast    = if CType ty == CType op
-#if MIN_VERSION_ghc(8,6,0)
                        then Just . EvExpr
-#else
-                       then Just
-#endif
-                       else makeKnCoercion cls ty op
+                       else makeKnCoercion cls ty op deps
       in  cast =<< lookup (CType kn) givens
 
     -- Find a known constraint for a wanted, so that (modulo normalization)
     -- the two are a constant offset apart.
-    offset :: Type -> TcPluginM (Maybe (EvTerm,[Ct]))
+    offset :: Type -> TcPluginM Solve (Maybe (EvTerm,[Ct]))
+    offset LitTy{} = pure Nothing
     offset want = runMaybeT $ do
       let -- Get the knownnat contraints
           unKn ty' = case classifyPredType ty' of
@@ -532,78 +501,104 @@ constraintToEvTerm defs givens (ct,cls,op,orig) = do
                          -> Just ty''
                        _ -> Nothing
           -- Get the rewrites
-          unEq ty' = case classifyPredType ty' of
-                       EqPred NomEq ty1 ty2 -> Just (ty1,ty2)
-                       _ -> Nothing
-          rewrites = mapMaybe (unEq . unCType . fst) givens
+          unEq (ty',ev) = case classifyPredType ty' of
+                            EqPred NomEq ty1 ty2 -> Just (ty1,ty2,ev)
+                            _ -> Nothing
+          rewrites :: [(Type,Type,EvExpr)]
+          rewrites = mapMaybe (unEq . first unCType) givens
           -- Rewrite
-          rewriteTy tyK (ty1,ty2) | ty1 `eqType` tyK = Just ty2
-                                  | ty2 `eqType` tyK = Just ty1
-                                  | otherwise        = Nothing
+          rewriteTy tyK (ty1,ty2,ev)
+            | ty1 `eqType` tyK
+            = Just (ty2,Just (tyK,evTermCoercion_maybe (EvExpr ev)))
+            | ty2 `eqType` tyK
+            = Just (ty1,Just (tyK,fmap mkSymCo (evTermCoercion_maybe (EvExpr ev))))
+            | otherwise
+            = Nothing
           -- Get only the [G]iven KnownNat constraints
           knowns   = mapMaybe (unKn . unCType . fst) givens
           -- Get all the rewritten KNs
           knownsR  = catMaybes $ concatMap (\t -> map (rewriteTy t) rewrites) knowns
-          knownsX  = knowns ++ knownsR
+          knownsX :: [(Type, Maybe (Type, Maybe Coercion))]
+          knownsX  = fmap (,Nothing) knowns ++ knownsR
           -- pair up the sum-of-products KnownNat constraints
           -- with the original Nat operation
           subWant  = mkTyConApp typeNatSubTyCon . (:[want])
-          exploded = map (fst . runWriter . normaliseNat . subWant &&& id)
+          -- exploded :: [()]
+          exploded = map (fst . runWriter . normaliseNat . subWant . fst &&& id)
                          knownsX
           -- interesting cases for us are those where
           -- wanted and given only differ by a constant
-          examineDiff (S [P [I n]]) entire = Just (entire,I n)
-          examineDiff (S [P [V v]]) entire = Just (entire,V v)
+          examineDiff ((S [P [I n]]),deps) entire = Just (entire,I n,deps)
+          examineDiff ((S [P [V v]]),deps) entire = Just (entire,V v,deps)
           examineDiff _ _ = Nothing
           interesting = mapMaybe (uncurry examineDiff) exploded
       -- convert the first suitable evidence
-      ((h,corr):_) <- pure interesting
+      (((h,sM),corr,deps):_) <- pure interesting
       x <- case corr of
-                I 0 -> pure h
+                I 0 -> pure (fromMaybe (h,Nothing) sM)
                 I i | i < 0
-                    -> pure (mkTyConApp typeNatAddTyCon [h,mkNumLitTy (negate i)])
+                    , let l1 = mkNumLitTy (negate i)
+                    -> case sM of
+                        Just (q,cM) -> pure
+                          ( mkTyConApp typeNatAddTyCon [q,l1]
+                          , fmap (mkTyConAppCo Nominal typeNatAddTyCon . (:[mkNomReflCo l1])) cM
+                          )
+                        Nothing -> pure
+                          ( mkTyConApp typeNatAddTyCon [h,l1]
+                          , Nothing
+                          )
                     | otherwise
-                    -> pure (mkTyConApp typeNatSubTyCon [h,mkNumLitTy i])
+                    , let l1 = mkNumLitTy i
+                    -> case sM of
+                        Just (q,cM) -> pure
+                          ( mkTyConApp typeNatSubTyCon [q,l1]
+                          , fmap (mkTyConAppCo Nominal typeNatSubTyCon . (:[mkNomReflCo l1])) cM
+                          )
+                        Nothing -> pure
+                          ( mkTyConApp typeNatSubTyCon [h,l1]
+                          , Nothing
+                          )
                 -- If the offset between a given and a wanted is again the wanted
                 -- then the given is twice the wanted; so we can just divide
                 -- the given by two. Only possible in GHC 8.4+; for 8.2 we simply
                 -- fail because we don't know how to divide.
-                c   | CType (reifySOP (S [P [c]])) == CType want ->
-#if MIN_VERSION_ghc(8,4,0)
-                     pure (mkTyConApp typeNatDivTyCon [h,reifySOP (S [P [I 2]])])
-#else
-                     MaybeT (pure Nothing)
-#endif
+                c   | CType (reifySOP (S [P [c]])) == CType want
+                    , let l2 = mkNumLitTy 2
+                    -> case sM of
+                        Just (q,cM) -> pure
+                          ( mkTyConApp typeNatDivTyCon [q,l2]
+                          , fmap (mkTyConAppCo Nominal typeNatDivTyCon . (:[mkNomReflCo l2])) cM
+                          )
+                        Nothing -> pure
+                          ( mkTyConApp typeNatDivTyCon [h,l2]
+                          , Nothing
+                          )
                 -- Only solve with a variable offset if we have [G]iven knownnat for it
                 -- Failing to do this check results in #30
-                V v | all (not . eqType (TyVarTy v)) knownsX
-                    -> MaybeT (pure Nothing)
-                _ -> pure (mkTyConApp typeNatSubTyCon [h,reifySOP (S [P [corr]])])
-      MaybeT (go x)
+                V v  | all (not . eqType (TyVarTy v) . fst) knownsX
+                     -> MaybeT (pure Nothing)
+                _    -> let lC = reifySOP (S [P [corr]]) in
+                        case sM of
+                          Just (q,cM) -> pure
+                            ( mkTyConApp typeNatSubTyCon [q,lC]
+                            , fmap (mkTyConAppCo Nominal typeNatSubTyCon . (:[mkNomReflCo lC])) cM
+                            )
+                          Nothing -> pure
+                            ( mkTyConApp typeNatSubTyCon [h,lC]
+                            , Nothing
+                            )
+      MaybeT (go deps x)
 
 makeWantedEv
   :: Ct
   -> Type
-#if MIN_VERSION_ghc(8,5,0)
-  -> TcPluginM (EvExpr,Ct)
-#else
-  -> TcPluginM (EvTerm,Ct)
-#endif
+  -> TcPluginM Solve (EvExpr,Ct)
 makeWantedEv ct ty = do
   -- Create a new wanted constraint
   wantedCtEv <- newWanted (ctLoc ct) ty
-#if MIN_VERSION_ghc(8,5,0)
   let ev      = ctEvExpr wantedCtEv
-#else
-  let ev      = ctEvTerm wantedCtEv
-#endif
       wanted  = mkNonCanonical wantedCtEv
-      -- Set the source-location of the new wanted constraint to the source
-      -- location of the [W]anted constraint we are currently trying to solve
-      ct_ls   = ctLocSpan (ctLoc ct)
-      ctl     = ctEvLoc  wantedCtEv
-      wanted' = setCtLoc wanted (setCtLocSpan ctl ct_ls)
-  return (ev,wanted')
+  return (ev,wanted)
 
 {- |
 Given:
@@ -623,6 +618,7 @@ dictionary for binary functions, the coercion happens in the following steps:
 
 this process is mirrored for the dictionary functions of a higher arity
 -}
+
 makeOpDict
   :: (Class,DFunId)
   -- ^ "magic" class function and dictionary function id
@@ -632,46 +628,34 @@ makeOpDict
   -- ^ Argument types for the Class
   -> [Type]
   -- ^ Argument types for the Instance
-  -> Type           -- ^ Type of the result
-#if MIN_VERSION_ghc(8,5,0)
+  -> Type
+  -- ^ Type of the result
+  -> [Coercion]
+  -- ^ Dependent coercions
   -> [EvExpr]
-#else
-  -> [EvTerm]
-#endif
   -- ^ Evidence arguments
+  -> Maybe (Type, Coercion)
   -> Maybe EvTerm
-makeOpDict (opCls,dfid) knCls tyArgsC tyArgsI z evArgs
-  | Just (_, kn_co_dict) <- tcInstNewTyCon_maybe (classTyCon knCls) [z]
-    -- KnownNat n ~ SNat n
-  , [ kn_meth ] <- classMethods knCls
-  , Just kn_tcRep <- tyConAppTyCon_maybe -- SNat
-                      $ funResultTy      -- SNat n
-                      $ dropForAlls      -- KnownNat n => SNat n
-                      $ idType kn_meth   -- forall n. KnownNat n => SNat n
-  , Just (_, kn_co_rep) <- tcInstNewTyCon_maybe kn_tcRep [z]
-    -- SNat n ~ Integer
-  , Just (_, op_co_dict) <- tcInstNewTyCon_maybe (classTyCon opCls) tyArgsC
-    -- KnownNatAdd a b ~ SNatKn (a+b)
-  , [ op_meth ] <- classMethods opCls
-  , Just (op_tcRep,op_args) <- splitTyConApp_maybe        -- (SNatKn, [KnownNatF2 f x y])
-                                 $ funResultTy            -- SNatKn (KnownNatF2 f x y)
-                                 $ (`piResultTys` tyArgsC) -- KnownNatAdd f x y => SNatKn (KnownNatF2 f x y)
-                                 $ idType op_meth         -- forall f a b . KnownNat2 f a b => SNatKn (KnownNatF2 f a b)
-  , Just (_, op_co_rep) <- tcInstNewTyCon_maybe op_tcRep op_args
+makeOpDict (opCls,dfid) knCls tyArgsC tyArgsI z deps evArgs sM
+  | let z1 = maybe z fst sM
     -- SNatKn (a+b) ~ Integer
-#if MIN_VERSION_ghc(8,5,0)
-  , let EvExpr dfun_inst = evDFunApp dfid tyArgsI evArgs
-#else
-  , let dfun_inst = EvDFunApp dfid tyArgsI evArgs
-#endif
+  , let dfun_inst = evDFunApp dfid tyArgsI evArgs
         -- KnownNatAdd a b
-        op_to_kn  = mkTcTransCo (mkTcTransCo op_co_dict op_co_rep)
-                                (mkTcSymCo (mkTcTransCo kn_co_dict kn_co_rep))
+  , let op_to_kn :: EvExpr -> EvExpr
+        op_to_kn ev
+            = wrapUnaryClassByFiat knCls [z1] deps
+            $ unwrapUnaryClassOverNewtype opCls tyArgsC ev
         -- KnownNatAdd a b ~ KnownNat (a+b)
-        ev_tm     = mkEvCast dfun_inst op_to_kn
-  = Just ev_tm
-  | otherwise
-  = Nothing
+  , let op_to_kn1 ev = case sM of
+          Nothing -> op_to_kn ev
+          Just (_,rw) ->
+            let kn_co_rw = mkTyConAppCo Representational (classTyCon knCls) [rw]
+                kn_co_co = mkPluginUnivCo "ghc-typelits-knownnat" Representational
+                              deps
+                              (coercionRKind kn_co_rw)
+                              (mkTyConApp (classTyCon knCls) [z])
+              in mkCast (op_to_kn ev) (mkTransCo kn_co_rw kn_co_co)
+  = Just $ EvExpr $ op_to_kn1 dfun_inst
 
 {-
 Given:
@@ -689,29 +673,13 @@ The coercion happens in the following steps:
 makeKnCoercion :: Class          -- ^ KnownNat class
                -> Type           -- ^ Type of the argument
                -> Type           -- ^ Type of the result
-#if MIN_VERSION_ghc(8,5,0)
+               -> [Coercion]     -- ^ Dependent coercions
                -> EvExpr
-#else
-               -> EvTerm
-#endif
                -- ^ KnownNat dictionary for the argument
                -> Maybe EvTerm
-makeKnCoercion knCls x z xEv
-  | Just (_, kn_co_dict_z) <- tcInstNewTyCon_maybe (classTyCon knCls) [z]
-    -- KnownNat z ~ SNat z
-  , [ kn_meth ] <- classMethods knCls
-  , Just kn_tcRep <- tyConAppTyCon_maybe -- SNat
-                      $ funResultTy      -- SNat n
-                      $ dropForAlls      -- KnownNat n => SNat n
-                      $ idType kn_meth   -- forall n. KnownNat n => SNat n
-  , Just (_, kn_co_rep_z) <- tcInstNewTyCon_maybe kn_tcRep [z]
-    -- SNat z ~ Integer
-  , Just (_, kn_co_rep_x) <- tcInstNewTyCon_maybe kn_tcRep [x]
-    -- Integer ~ SNat x
-  , Just (_, kn_co_dict_x) <- tcInstNewTyCon_maybe (classTyCon knCls) [x]
-    -- SNat x ~ KnownNat x
-  = Just . mkEvCast xEv $ (kn_co_dict_x `mkTcTransCo` kn_co_rep_x) `mkTcTransCo` mkTcSymCo (kn_co_dict_z `mkTcTransCo` kn_co_rep_z)
-  | otherwise = Nothing
+makeKnCoercion knCls x z deps knownNat_x
+  = Just $ EvExpr $ wrapUnaryClassByFiat knCls [z] deps
+                  $ unwrapUnaryClassOverNewtype knCls [x] knownNat_x
 
 -- | THIS CODE IS COPIED FROM:
 -- https://github.com/ghc/ghc/blob/8035d1a5dc7290e8d3d61446ee4861e0b460214e/compiler/typecheck/TcInteract.hs#L1973
@@ -722,34 +690,18 @@ makeKnCoercion knCls x z xEv
 --
 --     Integer -> SNat n     -- representation of literal to singleton
 --     SNat n  -> KnownNat n -- singleton to dictionary
-#if MIN_VERSION_ghc(8,5,0)
-makeLitDict :: Class -> Type -> Integer -> TcPluginM (Maybe EvTerm)
-#else
-makeLitDict :: Class -> Type -> Integer -> Maybe EvTerm
-#endif
-makeLitDict clas ty i
-  | Just (_, co_dict) <- tcInstNewTyCon_maybe (classTyCon clas) [ty]
-    -- co_dict :: KnownNat n ~ SNat n
-  , [ meth ]   <- classMethods clas
-  , Just tcRep <- tyConAppTyCon_maybe -- SNat
-                    $ funResultTy     -- SNat n
-                    $ dropForAlls     -- KnownNat n => SNat n
-                    $ idType meth     -- forall n. KnownNat n => SNat n
-  , Just (_, co_rep) <- tcInstNewTyCon_maybe tcRep [ty]
-        -- SNat n ~ Integer
-#if MIN_VERSION_ghc(8,5,0)
+makeLitDict :: Class
+            -> Type
+            -> [Coercion]
+                 -- ^ dependent coercions
+            -> Integer
+            -> TcPluginM Solve (Maybe EvTerm)
+makeLitDict clas ty deps i
   = do
-    et <- unsafeTcPluginTcM (mkNaturalExpr i)
-    let ev_tm = mkEvCast et (mkTcSymCo (mkTcTransCo co_dict co_rep))
-    return (Just ev_tm)
-  | otherwise
-  = return Nothing
-#else
-  , let ev_tm = mkEvCast (EvLit (EvNum i)) (mkTcSymCo (mkTcTransCo co_dict co_rep))
-  = Just ev_tm
-  | otherwise
-  = Nothing
-#endif
+    et <- mkNaturalExpr i
+    let
+      ev_tm = wrapUnaryClassByFiat clas [ty] deps et
+    return (Just $ EvExpr ev_tm)
 
 {- |
 Given:
@@ -781,43 +733,54 @@ makeOpDictByFiat
   -- ^ Argument types for the Instance
   -> Type
   -- ^ Type of the result
-#if MIN_VERSION_ghc(8,6,0)
+  -> [Coercion]
+  -- ^ Dependent coercions
   -> [EvExpr]
-#else
-  -> [EvTerm]
-#endif
   -- ^ Evidence arguments
   -> Maybe EvTerm
-#if MIN_VERSION_ghc(8,6,0)
-makeOpDictByFiat (opCls,dfid) knCls tyArgsC tyArgsI z evArgs
-    -- KnownBool b ~ SBool b
-  | Just (_, kn_co_dict) <- tcInstNewTyCon_maybe (classTyCon knCls) [z]
-  , [ kn_meth ] <- classMethods knCls
-  , Just kn_tcRep <- tyConAppTyCon_maybe -- SBool
-                       $ funResultTy     -- SBool b
-                       $ dropForAlls     -- KnownBool b => SBool b
-                       $ idType kn_meth  -- forall b. KnownBool b => SBool b
-    -- SBool b R~ Bool (The "Lie")
-  , let kn_co_rep = mkUnivCo (PluginProv "ghc-typelits-knownnat")
-                             Representational
-                             (mkTyConApp kn_tcRep [z]) boolTy
-    -- KnownBoolNat2 f a b ~ SBool f
-  , Just (_, op_co_dict) <- tcInstNewTyCon_maybe (classTyCon opCls) tyArgsC
-  , [ op_meth ] <- classMethods opCls
-  , Just (op_tcRep,op_args) <- splitTyConApp_maybe        -- (SBool, [f])
-                                 $ funResultTy            -- SBool f
-                                 $ (`piResultTys` tyArgsC) -- KnownBoolNat2 f x y => SBool f
-                                 $ idType op_meth         -- forall f x y . KnownBoolNat2 f a b => SBoolf f
-    -- SBoolF f ~ Bool
-  , Just (_, op_co_rep) <- tcInstNewTyCon_maybe op_tcRep op_args
-  , EvExpr dfun_inst <- evDFunApp dfid tyArgsI evArgs
-    -- KnownBoolNat2 f x y ~ KnownBool b
-  , let op_to_kn  = mkTcTransCo (mkTcTransCo op_co_dict op_co_rep)
-                                (mkTcSymCo (mkTcTransCo kn_co_dict kn_co_rep))
-        ev_tm     = mkEvCast dfun_inst op_to_kn
-  = Just ev_tm
+makeOpDictByFiat (opCls,dfid) knCls tyArgsC tyArgsI z deps evArgs
+  = Just $ EvExpr $ wrapUnaryClassByFiat knCls [z] deps
+                  $ unwrapUnaryClassOverNewtype opCls tyArgsC ev0
+  where
+    ev0 = evDFunApp dfid tyArgsI evArgs
+
+-- | Given a class of the form @class C a b c where { meth :: ... }@ with
+-- a single method, construct a dictionary of the class using an 'UnivCo'.
+wrapUnaryClassByFiat :: HasDebugCallStack => Class -> [Type] -> [Coercion] -> EvExpr -> EvExpr
+wrapUnaryClassByFiat cls tys deps et
+  | Just dc <- tyConSingleDataCon_maybe (classTyCon cls)
+  , [meth] <- classMethods cls
+  , let meth_ty = subst $ classMethodTy meth
+  = let
+      by_fiat =
+        mkPluginUnivCo "ghc-typelits-knownnat" Representational
+          deps
+          (exprType et)
+          meth_ty
+    in
+      Var (dataConWrapId dc) `mkTyApps` tys `mkApps` [mkCast et by_fiat]
   | otherwise
-  = Nothing
-#else
-makeOpDictByFiat _ _ _ _ _ _ = Nothing
-#endif
+  = pprPanic "wrapUnaryClassByFiat: class not of expected form" $
+      vcat [ text "cls:" <+> ppr cls
+           , text "tys:" <+> ppr tys
+           ]
+
+  where
+    subst = substTyWithUnchecked (classTyVars cls) tys
+
+-- | Given a class of the form @class C a b c where { meth :: N x y }@
+-- in which @N@ is a newtype, and a dictionary for this class, unwraps **both**
+-- the class and the newtype to obtain the value inside the newtype.
+unwrapUnaryClassOverNewtype :: HasDebugCallStack => Class -> [Type] -> EvExpr -> EvExpr
+unwrapUnaryClassOverNewtype cls tys et
+  | [sel] <- classMethods cls
+  , Just (rep_tc, rep_args) <- splitTyConApp_maybe (subst $ classMethodTy sel)
+  , Just (_, co) <- instNewTyCon_maybe rep_tc rep_args
+  = mkCast (evSelector sel tys [et]) co
+  | otherwise
+  = pprPanic "unwrapUnaryClassOverNewtype: class not of expected form" $
+      vcat [ text "cls:" <+> ppr cls
+           , text "tys:" <+> ppr tys
+           ]
+  where
+    subst = substTyWithUnchecked (classTyVars cls) tys
